# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from airflow.triggers.base import BaseEventTrigger


class BaseMessageQueueProvider:
    """
    Base class defining a provider supported by operators/triggers of common-messaging provider.

    To add a new provider supported by the provider, create a new class extending this base class and add it
    to ``MESSAGE_QUEUE_PROVIDERS``.
    """

    @abstractmethod
    def queue_matches(self, queue: str) -> bool:
        """
        Return whether a given queue (string) matches a specific provider's pattern.

        This function must be as specific as possible to avoid collision with other providers.
        Functions in this provider should NOT overlap with each other in their matching criteria.

        :param queue: The queue identifier
        """

    @abstractmethod
    def trigger_class(self) -> type[BaseEventTrigger]:
        """Trigger class to use when ``queue_matches`` returns True."""

    @abstractmethod
    def trigger_kwargs(self, queue: str, **kwargs) -> dict:
        """
        Parameters passed to the instance of ``trigger_class``.

        :param queue: The queue identifier
        """



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import importlib
from collections.abc import AsyncIterator
from functools import cached_property
from typing import Any

from airflow.providers_manager import ProvidersManager
from airflow.triggers.base import BaseEventTrigger, TriggerEvent

providers_manager = ProvidersManager()
providers_manager.initialize_providers_queues()


def create_class_by_name(name: str):
    module_name, class_name = name.rsplit(".", 1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


MESSAGE_QUEUE_PROVIDERS = [create_class_by_name(name)() for name in providers_manager.queue_class_names]


class MessageQueueTrigger(BaseEventTrigger):
    """
    ``MessageQueueTrigger`` serves as a unified trigger for monitoring message queues from different providers.

    It abstracts away provider-specific details, allowing users to monitor a queue with a single trigger,
    regardless of the underlying provider.

    This makes it easy to switch providers without modifying the trigger.

    :param queue: The queue identifier

    .. seealso::
        For more information on how to use this trigger, take a look at the guide:
        :ref:`howto/trigger:MessageQueueTrigger`
    """

    def __init__(self, *, queue: str, **kwargs: Any) -> None:
        self.queue = queue
        self.kwargs = kwargs

    @cached_property
    def trigger(self) -> BaseEventTrigger:
        if len(MESSAGE_QUEUE_PROVIDERS) == 0:
            self.log.error(
                "No message queue providers are available. "
                "Please ensure that you have the necessary providers installed."
            )
            raise ValueError("No message queue providers are available. ")
        providers = [provider for provider in MESSAGE_QUEUE_PROVIDERS if provider.queue_matches(self.queue)]
        if len(providers) == 0:
            self.log.error(
                "The queue '%s' is not recognized by any of the registered providers. "
                "The available providers are: '%s'.",
                self.queue,
                ", ".join([provider for provider in MESSAGE_QUEUE_PROVIDERS]),
            )
            raise ValueError("The queue is not recognized by any of the registered providers.")
        if len(providers) > 1:
            self.log.error(
                "The queue '%s' is recognized by more than one provider. "
                "At least two providers in ``MESSAGE_QUEUE_PROVIDERS`` are colliding with each "
                "other: '%s'",
                self.queue,
                ", ".join([provider for provider in providers]),
            )
            raise ValueError(f"The queue '{self.queue}' is recognized by more than one provider.")
        return providers[0].trigger_class()(
            **providers[0].trigger_kwargs(self.queue, **self.kwargs), **self.kwargs
        )

    def serialize(self) -> tuple[str, dict[str, Any]]:
        return self.trigger.serialize()

    async def run(self) -> AsyncIterator[TriggerEvent]:
        async for event in self.trigger.run():
            yield event



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

# [START howto_trigger_message_queue]
from airflow.providers.common.messaging.triggers.msg_queue import MessageQueueTrigger
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.sdk import DAG, Asset, AssetWatcher

# Define a trigger that listens to an external message queue (AWS SQS in this case)
trigger = MessageQueueTrigger(queue="https://sqs.us-east-1.amazonaws.com/0123456789/my-queue")

# Define an asset that watches for messages on the queue
asset = Asset("sqs_queue_asset", watchers=[AssetWatcher(name="sqs_watcher", trigger=trigger)])

with DAG(dag_id="example_msgq_watcher", schedule=[asset]) as dag:
    EmptyOperator(task_id="task")
# [END howto_trigger_message_queue]


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)
